{-

     Fun with the λ-calculus!


   Types for Proofs and Programs

            Ulf Norell

-}













module Term where

open import Prelude
open import Type

Name = String

-- λ-terms with natural numbers
data RawTerm : Set where
  var : (x : Name) → RawTerm
  lit : (n : Nat) → RawTerm
  suc : RawTerm
  app : (e e₁ : RawTerm) → RawTerm
  lam : (x : Name) → (a : Type) → (e : RawTerm) → RawTerm

-- λ (x : nat) . suc x
ex₁ : RawTerm
ex₁ = lam "x" nat (app suc (var "x"))

Context : Set
Context = List (Name × Type)

infix 1 _∈_
data _∈_ {A : Set} (x : A) : List A → Set where
  zero : ∀ {xs}            → x ∈ x ∷ xs
  suc  : ∀ {y xs} → x ∈ xs → x ∈ y ∷ xs

private
  2∈0123 : (2 ofType Nat) ∈ 0 ∷ 1 ∷ 2 ∷ 3 ∷ []
  2∈0123 = suc (suc zero)

data Term (Γ : Context) : Type → Set where
  lit : (n : Nat) → Term Γ nat
  suc : Term Γ (nat => nat)
  app : ∀ {σ τ} (s : Term Γ (σ => τ)) (t : Term Γ σ) → Term Γ τ
  lam : ∀ {σ τ} (x : Name) (t : Term ((x , σ) ∷ Γ) τ) → Term Γ (σ => τ)
  var : ∀ {τ} (x : Name) → (x , τ) ∈ Γ → Term Γ τ

forgetTypes : ∀ {Γ a} → Term Γ a → RawTerm
forgetTypes (lit n)           = lit n
forgetTypes suc               = suc
forgetTypes (app s t)         = app (forgetTypes s) (forgetTypes t)
forgetTypes (lam {σ = a} x t) = lam x a (forgetTypes t)
forgetTypes (var x i)         = var x
